"""Cython Backend
=================

Internal API
------------

.. autoclass:: HeaderFunction
   :members:
   :private-members:

.. autoclass:: SubBackendJITCython
   :members:
   :private-members:

.. autoclass:: CythonBackend
   :members:
   :private-members:

"""
import copy
import inspect

from warnings import warn

from transonic.analyses.extast import unparse, gast, FunctionDef, Name
from transonic.signatures import make_signatures_from_typehinted_func
from transonic.typing import format_type_as_backend_type, MemLayout

from .base import BackendAOT, TypeHintRemover, format_str
from .base_jit import SubBackendJIT
from .typing import TypeFormatter


def normalize_type_name_for_array(name):
    if name == "bool_":
        return "np.uint8"
    if any(name.endswith(str(number)) for number in (8, 16, 32, 64, 128)):
        return "np." + name
    if name in ("int", "float", "complex"):
        return "np." + name
    return name


class TypeFormatterCython(TypeFormatter):
    def normalize_type_name(self, name):
        if any(name.endswith(str(number)) for number in (8, 16, 32, 64, 128)):
            return "np." + name + "_t"
        if name in ("int", "float", "complex", "str"):
            return f"cython.{name}"
        return name

    def make_array_code(
        self, dtype, ndim, shape, memview, mem_layout, positive_indices
    ):
        dtype = normalize_type_name_for_array(dtype.__name__)
        if ndim == 0:
            return dtype

        if memview:
            return memoryview_type(dtype, ndim, mem_layout)
        else:
            return np_ndarray_type(dtype, ndim, mem_layout, positive_indices)

    def make_dict_code(self, type_keys, type_values, **kwargs):
        return "dict"

    def make_set_code(self, type_keys, **kwargs):
        return "set"

    def make_list_code(self, type_elem, **kwargs):
        return "list"

    def make_tuple_code(self, types, **kwargs):
        return "tuple"

    def make_const_code(self, code):
        return "const " + code


def memoryview_type(dtype, ndim, mem_layout) -> str:
    ndim_F = 0
    ndim_C = 0
    if mem_layout is MemLayout.C:
        ndim_C = 1
        ndim -= 1
    elif mem_layout is MemLayout.F:
        ndim_F = 1
        ndim -= 1
    end = ", ".join(["::1"] * ndim_F + [":"] * ndim + ["::1"] * ndim_C)
    return f"{dtype}_t[{end}]"


def np_ndarray_type(dtype, ndim, mem_layout, positive_indices) -> str:
    if mem_layout is MemLayout.C:
        mode = ', mode="c"'
    elif mem_layout is MemLayout.F:
        mode = ', mode="f"'
    else:
        mode = ""

    if positive_indices:
        positive_indices = ", negative_indices=False"
    else:
        positive_indices = ""

    return f"np.ndarray[{dtype}_t, ndim={ndim}{mode}{positive_indices}]"


class HeaderFunction:
    def __init__(
        self,
        path=None,
        name=None,
        arguments=None,
        types: dict = None,
        imports=None,
    ):

        if path is not None:
            self.path = path
            with open(path) as file:
                lines = file.readlines()

            last_line = lines[-1]
            assert last_line.startswith("cpdef ")
            name = last_line.split(" ", 1)[1].split("(", 1)[0]

            parts = [
                part.strip()
                for part in "".join(lines[:-1]).split("ctypedef fused ")
            ]
            imports = parts[0]

            types = {}

            for part in parts[1:]:
                assert part.startswith(f"__{name}_")
                lines = part.split("\n")
                arg_name = lines[0].split(f"__{name}_", 1)[1].split(":", 1)[0]
                types[arg_name] = set(line.strip() for line in lines[1:])

        if types is None:
            if arguments is None:
                raise ValueError
            types = {key: set() for key in arguments}

        if arguments is None:
            arguments = types.keys()

        self.arguments = arguments
        self.name = name
        self.types = types
        self.imports = imports

    def make_code(self):

        bits = [self.imports + "\n\n"]

        for arg, types in self.types.items():
            bits.append(f"ctypedef fused __{self.name}_{arg}:\n")
            for type_ in sorted(types):
                bits.append(f"    {type_}\n")
            bits.append("\n")

        tmp = ", ".join(f"__{self.name}_{arg} {arg}" for arg in self.types)
        bits.append(f"cpdef {self.name}({tmp})")
        code = "".join(bits)
        return code

    def add_signature(self, new_types):
        for new_type, set_types in zip(new_types, self.types.values()):
            set_types.add(new_type)

    def update_with_other_header(self, other):
        if self.name != other.name:
            raise ValueError
        if self.types.keys() != other.types.keys():
            raise ValueError
        for key, value in other.types.items():
            self.types[key].update(value)


class SubBackendJITCython(SubBackendJIT):
    def make_new_header(self, func, arg_types):
        # Include signature comming from type hints
        header = HeaderFunction(
            name=func.__name__,
            arguments=list(inspect.signature(func).parameters.keys()),
            imports="import cython\n\nimport numpy as np\ncimport numpy as np\n",
        )

        signatures = make_signatures_from_typehinted_func(
            func, self.type_formatter, as_list_str=True
        )

        for signature in signatures:
            header.add_signature(signature)

        if arg_types != "no types":
            header.add_signature(arg_types)

        return header

    def _load_old_header(self, path_backend_header):
        return HeaderFunction(path=path_backend_header)

    def _merge_header_objects(self, header, header_old):
        header.update_with_other_header(header_old)
        return header

    def _make_header_code(self, header):
        return header.make_code()


class CythonBackend(BackendAOT):
    """Main class for the Cython backend"""

    backend_name = "cython"
    suffix_header = ".pxd"
    keyword_export = "cpdef"
    _SubBackendJIT = SubBackendJITCython
    _TypeFormatter = TypeFormatterCython

    def _make_first_lines_header(self):
        return ["import cython\n\nimport numpy as np\ncimport numpy as np\n"]

    def _make_header_from_fdef_annotations(
        self, fdef, annotations: dict, locals_types=None, returns=None
    ):

        if hasattr(fdef, "_transonic_keywords"):
            decorator_keywords = fdef._transonic_keywords
        else:
            decorator_keywords = {}

        inline = decorator_keywords.get("inline", False)
        inline = "inline " if inline else ""

        fdef = FunctionDef(name=fdef.name, args=copy.deepcopy(fdef.args), body=[])

        assert isinstance(annotations, list)

        if len(annotations) > 1:
            warn(
                "Cython backend only supports one set of annotations. "
                "Please use Transonic fused types."
            )

        try:
            annotations = annotations[0]
        except IndexError:
            annotations = {}

        transonic_types = set(annotations.values())

        if locals_types:
            transonic_types.update(locals_types.values())

        if returns:
            transonic_types.add(returns)

        transonic_types = sorted(transonic_types, key=repr)

        template_parameters = set()
        for ttype in transonic_types:
            if hasattr(ttype, "get_template_parameters"):
                template_parameters.update(ttype.get_template_parameters())

        template_parameters = sorted(template_parameters, key=repr)

        transonic_fused_types = [
            ttype
            for ttype in transonic_types
            if hasattr(ttype, "is_fused_type") and ttype.is_fused_type()
        ]

        if not all(param.values for param in template_parameters):
            raise ValueError(
                f"{template_parameters}, {[param.values for param in template_parameters]}"
            )

        cython_fused_types = {}

        def get_ttype_name(ttype):
            if hasattr(ttype, "short_repr"):
                ttype_name = ttype.short_repr()
            elif hasattr(ttype, "__name__"):
                ttype_name = ttype.__name__
            elif isinstance(ttype, str):
                ttype_name = ttype
            else:
                raise RuntimeError
            return ttype_name

        for ttype in transonic_fused_types:
            ttype_name = get_ttype_name(ttype)
            name_cython_type = f"__{fdef.name}__{ttype_name}"

            cython_types = ttype.get_all_formatted_backend_types(
                self.type_formatter
            )
            if "None" in cython_types:
                cython_types.remove("None")
            cython_fused_types[name_cython_type] = cython_types

        signatures_func = []

        for name, possible_types in cython_fused_types.items():
            ctypedef = [f"ctypedef fused {name}:\n"]
            for possible_type in sorted(set(possible_types)):
                ctypedef.append(f"   {possible_type}\n")
            signatures_func.append("".join(ctypedef))

        def get_name_cython_type(ttype):
            ttype_name = get_ttype_name(ttype)
            name_cython_type = f"__{fdef.name}__{ttype_name}"
            if name_cython_type in cython_fused_types:
                return name_cython_type
            return format_type_as_backend_type(ttype, self.type_formatter)

        # change function parameters
        if fdef.args.defaults:
            name_start = Name("*", gast.Param())
            fdef.args.defaults = [name_start] * len(fdef.args.defaults)
        for name in fdef.args.args:
            name.annotation = None
            if annotations:
                ttype = annotations[name.id]
                name_cython_type = get_name_cython_type(ttype)
            else:
                name_cython_type = "object"
            name.id = f"{name_cython_type} {name.id}"

        if locals_types is not None and locals_types:
            # note: np.ndarray not supported by Cython in "locals"
            # TODO: thus, fused types not supported here
            locals_types = ", ".join(
                f"{k}={format_type_as_backend_type(v, self.type_formatter, memview=True)}"
                for k, v in locals_types.items()
            )
            signatures_func.append(f"@cython.locals({locals_types})")

        if returns is not None:
            ttype = returns
            name_cython_type = get_name_cython_type(ttype)
            returns = name_cython_type + " "
        else:
            returns = ""

        def_keyword = "cpdef"
        signatures_func.append(
            f"{def_keyword} {inline}{returns}{unparse(fdef).strip()[4:-1]}\n"
        )
        return signatures_func

    def _make_code_from_fdef_node(self, fdef):

        if hasattr(fdef, "_transonic_keywords"):
            decorator_keywords = fdef._transonic_keywords
        else:
            decorator_keywords = {}

        parts = []

        if not decorator_keywords.get("boundscheck", True):
            parts.append("@cython.boundscheck(False)")

        if not decorator_keywords.get("wraparound", True):
            parts.append("@cython.wraparound(False)")

        if decorator_keywords.get("cdivision", False):
            parts.append("@cython.cdivision(True)")

        if not decorator_keywords.get("nonecheck", True):
            parts.append("@cython.noneckeck(False)")

        if decorator_keywords.get("nogil", False):
            parts.append("@cython.nogil")

        transformed = TypeHintRemover().visit(fdef)
        # convert the AST back to source code
        parts.append(unparse(transformed))

        return format_str("\n".join(parts))

    def _make_beginning_code(self):
        return (
            "try:\n"
            "    import cython\n"
            "except ImportError:\n"
            "    from transonic_cl import cython\n\n"
        )



"""SubBackend class for Transonic JIT compilation
=================================================

Internal API
------------

.. autoclass:: SubBackendJIT
   :members:
   :private-members:

"""

import re
from pathlib import Path
from time import sleep

try:
    import numpy as np
except ImportError:
    np = None

from transonic.analyses import extast
from transonic.signatures import make_signatures_from_typehinted_func
from transonic.log import logger
from transonic import mpi
from transonic.typing import format_type_as_backend_type, typeof
from transonic.util import get_source_without_decorator, path_root

from .for_classes import produce_code_class


class SubBackendJIT:
    """Sub-class for Transonic JIT compilation"""

    def __init__(self, name, type_formatter):
        self.name = name
        self.name_capitalized = name.capitalize()
        self.type_formatter = type_formatter

        self.path_base = Path(path_root) / self.name / "__jit__"
        self.path_class = self.path_base.parent / "__jit_class__"

        if mpi.rank == 0:
            self.path_base.mkdir(parents=True, exist_ok=True)
            self.path_class.mkdir(parents=True, exist_ok=True)

        while True:
            if self.path_base.exists() and self.path_class.exists():
                break
            sleep(0.1)

    def make_backend_source(self, info_analysis, func, path_backend):
        func_name = func.__name__
        jitted_dicts = info_analysis["jitted_dicts"]
        src = info_analysis["codes_dependance"][func_name]
        if func_name in info_analysis["special"]:
            if func_name in jitted_dicts["functions"]:
                src += "\n" + extast.unparse(jitted_dicts["functions"][func_name])
            elif func_name in jitted_dicts["methods"]:
                src += "\n" + extast.unparse(jitted_dicts["methods"][func_name])
        else:
            # TODO find a prettier solution to remove decorator for cython
            # than doing two times a regex
            src += "\n" + re.sub(
                r"@.*?\sdef\s", "def ", get_source_without_decorator(func)
            )
        has_to_write = True
        if path_backend.exists() and mpi.rank == 0:
            with open(path_backend) as file:
                src_old = file.read()
            if src_old == src:
                has_to_write = False

        return src, has_to_write

    def make_new_header(self, func, arg_types):
        # Include signature comming from type hints
        signatures = make_signatures_from_typehinted_func(
            func, self.type_formatter
        )
        exports = set(f"export {signature}" for signature in signatures)

        if arg_types != "no types":
            exports.add(f"export {func.__name__}({', '.join(arg_types)})")
        return exports

    def merge_old_and_new_header(self, path_backend_header, header, func):

        try:
            path_backend_header_exists = path_backend_header.exists()
        except TimeoutError:
            raise RuntimeError(
                f"A MPI communication in Transonic failed when compiling "
                f"function {func}. This usually arises when a jitted "
                "function has to be compiled in MPI and is only called "
                f"by one process (rank={mpi.rank})."
            )

        if path_backend_header_exists:
            # get the old signature(s)
            header_old = self._load_old_header(path_backend_header)
            # FIXME: what do we do with the old signatures?
            header = self._merge_header_objects(header, header_old)

        return self._make_header_code(header)

    def _load_old_header(self, path_backend_header):
        exports_old = None
        if mpi.rank == 0:
            with open(path_backend_header) as file:
                exports_old = [export.strip() for export in file.readlines()]
        exports_old = mpi.bcast(exports_old)
        return exports_old

    def _merge_header_objects(self, header, header_old):
        header.update(header_old)
        return header

    def _make_header_code(self, header):
        return "\n".join(sorted(header)) + "\n"

    def write_new_header(self, path_backend_header, header, arg_types):
        mpi.barrier()
        if mpi.rank == 0:
            logger.debug(
                f"write {self.name_capitalized} signature in file "
                f"{path_backend_header} with types\n{arg_types}"
            )
            with open(path_backend_header, "w") as file:
                file.write(header)
                file.flush()

    def compute_typename_from_object(self, obj: object):
        """return the backend type name"""
        transonic_type = typeof(obj)
        return format_type_as_backend_type(transonic_type, self.type_formatter)

    def produce_code_class(self, cls):
        return produce_code_class(cls)



"""Code analyses
================

.. autosummary::
   :toctree:

    blocks_if
    capturex
    extast
    justintime
    parser
    util
    objects_from_str

"""

import re
from pprint import pformat

import gast as ast
from transonic.analyses import beniget

from transonic.log import logger

from .util import (
    filter_code_typevars,
    get_annotations,
    print_dumped,
    print_unparsed,
    find_path,
    get_exterior_code,
    extract_variable_annotations,
    extract_returns_annotation,
)
from .capturex import CaptureX
from .blocks_if import get_block_definitions
from .parser import parse_transonic_def_commands
from .objects_from_str import replace_strings_by_objects
from . import extast


__all__ = ["print_dumped", "print_unparsed"]


def compute_ancestors_chains(module_node):
    """Create Beniget objects"""

    ancestors = beniget.Ancestors()
    ancestors.visit(module_node)

    duc = beniget.DefUseChains()
    duc.visit(module_node)

    udc = beniget.UseDefChains(duc)

    return ancestors, duc, udc


def find_decorated_function(module, function_name: str, pathfile: str = None):
    ext_module = False
    for node in module.body:
        if isinstance(node, ast.FunctionDef):
            if node.name == function_name:
                return node, ext_module
        # look for function_name in the imports of module
        if isinstance(node, ast.ImportFrom):
            for func in node.names:
                if func.name == function_name:
                    # find and read the imported module file
                    name, path = find_path(node, pathfile)
                    with open(path) as file:
                        ext_module = extast.parse(file.read())
                    # find the definition of function_name in the imported module
                    node, _ = find_decorated_function(ext_module, function_name)
                    return node, ext_module
    raise RuntimeError


def get_decorated_dicts(
    module, ancestors, duc, pathfile: str, backend_name, decorator="boost"
):
    """Get the definitions of the decorated functions and classes"""

    kinds = ("functions", "functions_ext", "methods", "classes")

    backend_names = ("pythran", "cython", "numba", "python")
    decorated_dicts = {
        kind: {name: {} for name in backend_names} for kind in kinds
    }

    def add_definition(node_using_decorator):
        """

        NotImplemented:

        decor = boost(inline=1)

        isinstance(node_using_decorator, ast.Call)
        isinstance(node_parent, ast.Assign)
        node_parent.value.keywords

        """

        backend_name_def = backend_name
        decorated_dict = None
        ext_module = False
        node_parent = ancestors.parent(node_using_decorator)

        # first, get the keywords
        keywords = None
        if isinstance(node_using_decorator, ast.Call):
            # decorator used in the form boost(...)
            if node_using_decorator.keywords:
                # decorator used in the form boost(key=value)
                keywords = {
                    keyword.arg: eval(extast.unparse(keyword.value))
                    for keyword in node_using_decorator.keywords
                }
                backend_name_def = keywords.get("backend", backend_name)

        if isinstance(
            node_using_decorator, (ast.FunctionDef, ast.ClassDef)
        ) and isinstance(node_parent, (ast.Module, ast.ClassDef)):
            """case

            @boost
            def f():
                ...

            """
            definition_node = node_using_decorator

        elif isinstance(node_using_decorator, ast.Call) and isinstance(
            node_parent, (ast.FunctionDef, ast.ClassDef)
        ):
            """
            @boost(inline=1)
            def f():
                pass
            """

            definition_node = node_parent

        elif isinstance(node_using_decorator, ast.Call) and isinstance(
            node_parent, ast.Assign
        ):
            """
            f1_ = boost(f1)
            or
            decor = boost(foo=True)
            """

            call = node_using_decorator
            if call.keywords:
                "decor = boost(foo=True)"
                raise NotImplementedError
            else:
                "f1_ = boost(f1)"
                if len(call.args) > 1:
                    raise NotImplementedError
                func_name = call.args[0].id

                definition_node, ext_module = find_decorated_function(
                    module, func_name, pathfile
                )

        elif isinstance(node_using_decorator, ast.Call) and isinstance(
            node_parent, ast.Call
        ):
            """f1_ = boost(inline=1)(f1)"""

            call = node_parent
            if call.keywords or len(call.args) > 1:
                raise NotImplementedError
            func_name = call.args[0].id
            definition_node, ext_module = find_decorated_function(
                module, func_name, pathfile
            )

        else:
            raise NotImplementedError

        if keywords is not None:
            definition_node._transonic_keywords = keywords

        # if the definition node is in an imported module
        if ext_module:
            decorated_dict = decorated_dicts["functions_ext"][backend_name_def]
            decorated_dict[func_name] = (definition_node, ext_module)
        # If the definition node is not imported
        else:
            func_name = definition_node.name
            if isinstance(definition_node, ast.FunctionDef):
                parent = ancestors.parent(definition_node)
                if isinstance(parent, ast.ClassDef):
                    decorated_dict = decorated_dicts["methods"][backend_name_def]
                    func_name = (parent.name, func_name)
                else:
                    decorated_dict = decorated_dicts["functions"][
                        backend_name_def
                    ]
            elif isinstance(definition_node, ast.ClassDef):
                decorated_dict = decorated_dicts["classes"][backend_name_def]
            else:
                raise RuntimeError
            if decorated_dict is not None:
                decorated_dict[func_name] = definition_node

    # we first need to find the node where transonic.boost is defined...
    for node in module.body:
        if isinstance(node, ast.Import):
            for alias in node.names:
                if alias.name == "transonic":
                    transonic_def_node = alias
                    transonic_def = duc.chains[transonic_def_node]
                    for user in transonic_def.users():
                        for user1 in user.users():
                            if (
                                isinstance(user1.node, ast.Attribute)
                                and user1.node.attr == decorator
                            ):
                                definition_node = ancestors.parent(user1.node)
                                add_definition(definition_node)
        elif isinstance(node, ast.ImportFrom):
            if node.module == "transonic":
                for alias in node.names:
                    if alias.name == decorator:
                        boost_def_node = alias
                        boost_def = duc.chains[boost_def_node]
                        for user in boost_def.users():
                            definition_node = ancestors.parent(user.node)
                            add_definition(definition_node)

    return decorated_dicts


def get_types_from_transonic_signature(signature: str, function_name: str):
    types = []
    tmp = signature[len(function_name) + 1 : -1]

    while tmp:
        try:
            index_comma = tmp.index(",")
        except ValueError:
            tmp = tmp.strip()
            if tmp:
                types.append(tmp)
            break

        try:
            index_open_bracket = tmp.index("[")
        except ValueError:
            index_open_bracket = None

        if index_open_bracket is None or index_comma < index_open_bracket:
            type_, tmp = tmp.split(",", 1)

        else:
            index_close_bracket = tmp.index("]")
            # cover int[][] and int[:, :] but could be buggy!
            type0 = tmp[:index_close_bracket]
            type1, tmp = tmp[index_close_bracket:].split(",", 1)

            type_ = type0 + type1

        types.append(type_.strip())

    return types


def analyse_aot(code, pathfile, backend_name):
    """Gather the informations for ``@boost`` and blocks"""
    debug = logger.debug

    debug("extast.parse")
    module = extast.parse(code)

    debug("compute ancestors and chains")
    ancestors, duc, udc = compute_ancestors_chains(module)

    debug("filter_code_typevars")
    code_dependance_annotations = filter_code_typevars(module, duc, ancestors)
    debug(code_dependance_annotations)

    debug("find boosted objects")
    boosted_dicts = get_decorated_dicts(
        module, ancestors, duc, None, backend_name
    )
    debug(pformat(boosted_dicts))

    debug("compute the annotations")

    from transonic import aheadoftime

    aheadoftime.is_transpiling = True

    def_nodes = [
        def_node
        for boosted_dict in boosted_dicts.values()
        for boosted_dict_backend in boosted_dict.values()
        for def_node in boosted_dict_backend.values()
    ]

    capturex = CaptureX(
        def_nodes,
        module,
        ancestors,
        defuse_chains=duc,
        usedef_chains=udc,
        consider_annotations="only",
    )

    code_dependance_annotations = capturex.make_code_external()

    namespace = {}
    exec(code_dependance_annotations, namespace)

    aheadoftime.is_transpiling = False

    annotations = {}

    for kind, boosted_dict in boosted_dicts.items():
        annotations[kind] = {}
        for boosted_dict_backend in boosted_dict.values():
            for key, definition in boosted_dict_backend.items():
                ann = get_annotations(definition, namespace)
                if ann != {}:
                    annotations[kind][key] = ann

    debug(pformat(annotations))

    debug("get_block_definitions")
    blocks = get_block_definitions(code, module, ancestors, duc, udc)

    for block in blocks:
        replace_strings_by_objects(block.signatures, module, ancestors, udc, duc)

    debug(pformat(blocks))

    debug("compute code dependance:")

    # remove the decorator (boost) to compute the code dependance
    # + do not consider the class annotations
    for def_node in def_nodes:
        def_node.decorator_list = []
        if isinstance(def_node, ast.ClassDef):
            def_node.body = [
                node
                for node in def_node.body
                if not isinstance(node, ast.AnnAssign)
            ]

    blocks_for_capturex = [block.ast_code for block in blocks]

    capturex = CaptureX(
        def_nodes,
        module,
        ancestors=ancestors,
        defuse_chains=duc,
        usedef_chains=udc,
        consider_annotations=False,
        blocks=blocks_for_capturex,
    )
    code_ext = {"function": {}, "class": {}}
    code_dependance = capturex.make_code_external()
    # TODO implement class for new backends + debug this code :-)
    if boosted_dicts["functions"]["pythran"]:
        func = next(iter(boosted_dicts["functions"]["pythran"]))
        code_dependance, code_ext, _, _ = get_exterior_code(
            {func: code_dependance}, pathfile, classes="function", relative=True
        )
        code_dependance = code_dependance[func]
    if boosted_dicts["classes"]["pythran"]:
        cls = next(iter(boosted_dicts["classes"]["pythran"]))
        code_dependance, code_ext, _, _ = get_exterior_code(
            {cls: code_dependance}, pathfile, classes="class", relative=True
        )
        code_dependance = code_dependance[cls]
    debug(code_dependance)

    debug("parse_transonic_def_commands")
    signatures_p = parse_transonic_def_commands(code)

    annotations["__in_comments__"] = {}
    annotations["__locals__"] = {}
    annotations["__returns__"] = {}

    # match a comma except if inside backets (for `float[:,:]`)
    pattern = re.compile(r",\s*(?![^[]*\])")

    for functions_backend in boosted_dicts["functions"].values():
        for name_func, fdef in functions_backend.items():
            try:
                signatures = signatures_p[name_func]
            except KeyError:
                signatures = tuple()
            arg_names = [arg.id for arg in fdef.args.args]
            annotations_sign = []
            for sig in signatures:
                types = [
                    type_.strip()
                    for type_ in re.split(pattern, sig[len(fdef.name) + 1 : -1])
                ]
                annotations_sign.append(dict(zip(arg_names, types)))
            if annotations_sign:
                annotations["__in_comments__"][name_func] = annotations_sign

            # locals: variable annotations
            annotations_locals = extract_variable_annotations(fdef, namespace)
            if annotations_locals:
                annotations["__locals__"][name_func] = annotations_locals

            if fdef.returns:
                annotations["__returns__"][
                    name_func
                ] = extract_returns_annotation(fdef.returns, namespace)

    for signatures in annotations["__in_comments__"].values():
        replace_strings_by_objects(signatures, module, ancestors, udc, duc)

    debug("annotations:\n" + pformat(annotations))

    return boosted_dicts, code_dependance, annotations, blocks, code_ext



"""Use Pythran to create extensions
===================================

User API
--------

.. autofunction:: wait_for_all_extensions

Internal API
------------

.. autofunction:: make_hex

.. autoclass:: SchedulerPopen
   :members:
   :private-members:

.. autofunction:: compile_extension

"""

import multiprocessing
import subprocess
import time
from typing import Union, Optional
import sysconfig
import hashlib
import sys
import os
from datetime import datetime
import logging

from transonic import mpi
from transonic.mpi import Path, PathSeq
from transonic.log import logger
from transonic.progress import Progress

ext_suffix = sysconfig.get_config_var("EXT_SUFFIX") or ".so"


def modification_date(filename):
    """Get the modification date of a file"""
    return datetime.fromtimestamp(os.path.getmtime(str(filename)))


def has_to_build(output_file: Path, input_file: Path):
    """Check if a file has to be (re)built"""
    output_file = PathSeq(output_file)
    input_file = PathSeq(input_file)
    if not output_file.exists():
        return True
    mod_date_output = modification_date(output_file)
    if mod_date_output < modification_date(input_file):
        return True
    return False


def make_hex(src):
    """Produce a hash from a sting"""
    return hashlib.md5(src.encode("utf8")).hexdigest()


class SchedulerPopen:
    """Limit the number of compilations performed in parallel"""

    deltat = 0.2

    def __init__(self, parallel=True):
        self.progress = Progress(redirect_stdout=False, redirect_stderr=False)
        if mpi.rank > 0:
            return
        self.processes = []
        if parallel:
            self.limit_nb_processes = max(1, multiprocessing.cpu_count() // 2)
        else:
            self.limit_nb_processes = 1

    def block_until_avail(self, parallel=True):

        if mpi.rank == 0:
            if parallel:
                limit = self.limit_nb_processes
            else:
                limit = 1

            while len(self.processes) >= limit:
                time.sleep(self.deltat)
                self.processes = [
                    process
                    for process in self.processes
                    if process.is_alive_root()
                ]

        mpi.barrier(timeout=None)

    def wait_for_all_extensions(self):
        """Wait until all compilation processes are done"""
        if mpi.rank == 0:
            total = len(scheduler.processes)
            task = self.progress.add_task("Wait for all extensions", total=total)

            while self.processes:
                time.sleep(self.deltat)
                self.processes = [
                    process
                    for process in self.processes
                    if process.is_alive_root()
                ]
                self.progress.update(task, completed=total - len(self.processes))

        mpi.barrier(timeout=None)

    def compile_extension(
        self,
        path: Path,
        backend: str,
        name_ext_file: str,
        native=False,
        xsimd=False,
        openmp=False,
        str_accelerator_flags: Optional[str] = None,
        parallel=True,
        force=True,
    ):
        if not force:
            path_out = path.with_name(name_ext_file)
            if not has_to_build(path_out, path):
                logger.warning(
                    f"Do not {backend}ize {path} because it seems up-to-date "
                    "(but the compilation options may have changed). "
                    "You can force the compilation with the option -f."
                )
                return

        if mpi.rank == 0:
            task = self.progress.add_task(
                f"Schedule {backend}ization: {path.name}"
            )

        def advance(value):
            if mpi.rank == 0:
                self.progress.update(task, advance=value)

        if str_accelerator_flags is not None:
            flags = str_accelerator_flags.strip().split()
        else:
            flags = []

        def update_flags(flag):
            if flag not in flags:
                flags.append(flag)

        if native and os.name != "nt":
            update_flags("-march=native")

        if xsimd:
            update_flags("-DUSE_XSIMD")

        if openmp:
            update_flags("-fopenmp")

        if logger.is_enable_for("debug"):
            update_flags("-v")

        if logger.getEffectiveLevel() < logging.INFO:
            env = dict(os.environ, TRANSONIC_DEBUG="1")
        else:
            env = None

        words_command = [
            sys.executable,
            "-m",
            "transonic_cl.run_backend",
            path.name,
            "-b",
            backend,
        ]

        words_command.extend(("-o", name_ext_file))

        words_command.extend(flags)

        cwd = path.parent

        advance(10)
        self.block_until_avail(parallel)
        advance(20)

        process = None
        if mpi.rank == 0:
            if logger.getEffectiveLevel() <= logging.INFO:
                stdout = stderr = None
            else:
                stdout = stderr = subprocess.PIPE

            process = subprocess.Popen(
                words_command,
                cwd=cwd,
                stdout=stdout,
                stderr=stderr,
                universal_newlines=True,
                env=env,
            )

        process = mpi.ShellProcessMPI(process)

        if mpi.rank == 0:
            self.processes.append(process)

        advance(70)

        return process


scheduler = SchedulerPopen()


def wait_for_all_extensions():
    """Wait until all compilation processes are done"""
    scheduler.wait_for_all_extensions()


def compile_extension(
    path: Union[Path, str],
    backend: str,
    name_ext_file: str,
    native=False,
    xsimd=False,
    openmp=False,
    str_accelerator_flags: Optional[str] = None,
    parallel=False,
    force=False,
):
    if not isinstance(path, Path):
        path = Path(path)

    # return the process
    return scheduler.compile_extension(
        path,
        backend,
        name_ext_file,
        native=native,
        xsimd=xsimd,
        openmp=openmp,
        str_accelerator_flags=str_accelerator_flags,
        parallel=parallel,
        force=force,
    )



"""Compute function signatures
==============================

Internal API
------------

.. autofunction:: _format_types_as_backend_types

.. autofunction:: compute_signatures_from_typeobjects

.. autofunction:: _make_signature_from_template_variables

.. autofunction:: make_signatures_from_typehinted_func

"""

import itertools
import inspect
from typing import List

from transonic.typing import format_type_as_backend_type, str2type


def _format_types_as_backend_types(types, backend_type_formatter, **kwargs):
    """Compute a list of Pythran/Cython/... types

    """
    backend_types = []
    for type_ in types:
        backend_types.append(
            format_type_as_backend_type(type_, backend_type_formatter, **kwargs)
        )

    # TODO: handle this with an exception
    if "_empty" in backend_types:
        raise ValueError(
            "At least one annotation type lacking in a signature.\n"
            f"types = {types}"
        )

    return backend_types


def compute_signatures_from_typeobjects(
    types_in, backend_type_formatter
) -> List[List[str]]:
    """Compute a list of lists (signatures) of strings (backend types)

    """
    if isinstance(types_in, dict):
        types_in = types_in.values()

    types = []
    for type_ in types_in:
        if isinstance(type_, str):
            type_ = str2type(type_)
        types.append(type_)

    template_parameters = set()
    for type_ in types:
        if hasattr(type_, "get_template_parameters"):
            template_parameters.update(type_.get_template_parameters())

    if not template_parameters:
        if "_empty" in types:
            raise ValueError(
                "At least one annotation type lacking in a signature.\n"
                f"types = {types}"
            )
        str_types = [
            format_type_as_backend_type(type_, backend_type_formatter)
            for type_ in types
        ]
        return (str_types,)

    if not all(param.values for param in template_parameters):
        raise ValueError(
            f"{template_parameters}, {[param.values for param in template_parameters]}"
        )

    values_template_parameters = {}
    for param in template_parameters:
        values_template_parameters[param.__name__] = param.values

    backend_types = []
    names = values_template_parameters.keys()
    for set_types in itertools.product(*values_template_parameters.values()):
        template_variables = dict(zip(names, set_types))

        backend_types.append(
            _format_types_as_backend_types(
                types, backend_type_formatter, **template_variables
            )
        )

    return backend_types


def _make_signature_from_template_variables(
    func, backend_type_formatter, _signature=None, as_list_str=False, **kwargs
):
    """Create signature for a function with values for the template types

    (This function should only be used in
    :func:`make_signatures_from_typehinted_func`)

    Parameters
    ----------

    func: a function

    kwargs : dict

        The template types and their value

    """
    if _signature is None:
        _signature = inspect.signature(func)

    types = [param.annotation for param in _signature.parameters.values()]

    backend_types = _format_types_as_backend_types(
        types, backend_type_formatter, **kwargs
    )

    # "multiply" the signatures to take into account the "or" syntax
    multi_pythran_types = [
        _ for _ in itertools.product(*[t.split(" or ") for t in backend_types])
    ]
    signatures = []
    for backend_types in multi_pythran_types:
        if as_list_str:
            signature = backend_types
        else:
            signature = f"{func.__name__}(" + ", ".join(backend_types) + ")"
        signatures.append(signature)

    return signatures


def make_signatures_from_typehinted_func(
    func, backend_type_formatter, as_list_str=False
):
    """Make the signatures from annotations if it is possible

    Useful when there are only "not templated" types.

    """
    annotations = func.__annotations__

    if not annotations:
        return tuple()

    for key, value in annotations.items():
        if isinstance(value, str):
            annotations[key] = str2type(value)

    types = annotations.values()

    template_parameters = []
    for type_ in types:
        if hasattr(type_, "get_template_parameters"):
            template_parameters.extend(type_.get_template_parameters())
    template_parameters = set(template_parameters)

    _signature = inspect.signature(func)

    if not template_parameters:
        signatures = _make_signature_from_template_variables(
            func,
            backend_type_formatter,
            _signature=_signature,
            as_list_str=as_list_str,
        )
        return signatures

    if not all(param.values for param in template_parameters):
        return tuple()

    values_template_parameters = {}
    for param in template_parameters:
        values_template_parameters[param.__name__] = param.values

    names = values_template_parameters.keys()
    signatures = []
    for set_types in itertools.product(*values_template_parameters.values()):
        template_variables = dict(zip(names, set_types))
        signatures.extend(
            _make_signature_from_template_variables(
                func,
                backend_type_formatter,
                _signature=_signature,
                as_list_str=as_list_str,
                **template_variables,
            )
        )

    return signatures



print("import: 'transonic'")
import transonic

print("import: 'transonic.analyses'")
import transonic.analyses

print("import: 'transonic.backends'")
import transonic.backends

print("import: 'transonic_cl'")
import transonic_cl




"""Python backend
=================

Internal API
------------

.. autoclass:: SubBackendJITPython
   :members:
   :private-members:

.. autoclass:: PythonBackend
   :members:
   :private-members:

"""

from shutil import copyfile
from typing import Optional

from .base import BackendJIT
from .base_jit import SubBackendJIT


class SubBackendJITPython(SubBackendJIT):
    def make_new_header(self, func, arg_types):
        return ""

    def merge_old_and_new_header(self, path_backend_header, header, func):
        return ""

    def write_new_header(self, path_backend_header, header, arg_types):
        pass


class PythonBackend(BackendJIT):
    """Main class for the Python backend"""

    backend_name = "python"
    _SubBackendJIT = SubBackendJITPython

    def compile_extension(
        self,
        path_backend,
        name_ext_file=None,
        native=False,
        xsimd=False,
        openmp=False,
        str_accelerator_flags: Optional[str] = None,
        parallel=True,
        force=True,
    ):
        if name_ext_file is None:
            name_ext_file = self.name_ext_from_path_backend(path_backend)

        copyfile(path_backend, path_backend.with_name(name_ext_file))
        compiling = False
        process = None
        return compiling, process



from transonic._version import __version__

from transonic.aheadoftime import Transonic, boost
from transonic.backends import set_backend_for_this_module
from transonic.config import set_backend
from transonic.compiler import wait_for_all_extensions
from transonic.justintime import jit, set_compile_jit
from transonic.util import set_compile_at_import
from transonic.typing import (
    Array,
    NDim,
    Type,
    Union,
    List,
    Tuple,
    Dict,
    Set,
    str2type,
    typeof,
    Optional,
    const,
)


__all__ = [
    "__version__",
    "Transonic",
    "boost",
    "jit",
    "Array",
    "NDim",
    "Type",
    "List",
    "Dict",
    "Tuple",
    "Set",
    "Union",
    "Optional",
    "set_backend",
    "set_backend_for_this_module",
    "set_compile_jit",
    "set_compile_at_import",
    "str2type",
    "typeof",
    "wait_for_all_extensions",
]



"""Utilities for the setup.py files
===================================

User API
--------

Provides the classes PythranBuildExt and PythranExtension to be used in the
setup.py.

.. autofunction:: detect_transonic_extensions

.. autofunction:: init_transonic_extensions

"""

import os
import sys
from pathlib import Path
from sysconfig import get_config_var
from typing import Iterable
from concurrent.futures import ThreadPoolExecutor as Pool
import re

from setuptools.command.build_ext import build_ext as SetuptoolsBuildExt

try:
    from Cython.Distutils.build_ext import build_ext as CythonBuildExt
    from Cython.Build import cythonize
except ImportError:
    build_ext_classes = [SetuptoolsBuildExt]
    can_import_cython = False
else:
    build_ext_classes = [CythonBuildExt]
    can_import_cython = True

try:
    from pythran.dist import PythranBuildExt, PythranExtension
except ImportError:
    PythranBuildExt = object
    PythranExtension = object
    can_import_pythran = False
else:
    build_ext_classes.insert(0, PythranBuildExt)
    can_import_pythran = True

from transonic.util import modification_date
from transonic.config import backend_default
from transonic.backends import make_backend_files, backends
from transonic.util import can_import_accelerator
from transonic.log import get_logger

__all__ = [
    "PythranBuildExt",
    "PythranExtension",
    "can_import_pythran",
    "detect_transonic_extensions",
    "init_pythran_extensions",
    "init_transonic_extensions",
    "get_logger",
    "ParallelBuildExt",
    "make_backend_files",
]


def detect_transonic_extensions(
    name_package: str, backend: str = backend_default
) -> Iterable[str]:
    """Recursively scans a package for Pythran extensions to build, and returns a
    list of strings, where each string is a module name. The package should be
    present in the current working directory.

    """
    if backend != "numba" and not can_import_accelerator(backend):
        return []
    ext_names = []
    if not os.path.exists(str(name_package)):
        raise FileNotFoundError(f"Check the name of the package: {name_package}")

    backend = backends[backend]
    if backend.suffix_extension == ".py":
        # we have to filter out the "extensions"
        pattern = re.compile("_[a-f0-9]{32}.py$")

    extension = ".py"
    for root, dirs, files in os.walk(str(name_package)):
        path_dir = Path(root)
        for name in files:
            if (
                path_dir.name == f"__{backend.name}__"
                and name.endswith(extension)
                and not name.startswith("__ext__")
            ):
                path = path_dir / name
                if (
                    backend.suffix_extension == ".py"
                    and pattern.search(name) is not None
                ):
                    continue
                ext_names.append(
                    str(path).replace(os.path.sep, ".").split(extension)[0]
                )

    return ext_names


def init_pythran_extensions(
    name_package: str,
    include_dirs: Iterable[str] = (),
    compile_args: Iterable[str] = (),
    exclude_exts: Iterable[str] = (),
    logger=None,
):
    return init_transonic_extensions(
        name_package, "pythran", include_dirs, compile_args, exclude_exts, logger
    )


def init_transonic_extensions(
    name_package: str,
    backend: str = backend_default,
    include_dirs: Iterable[str] = (),
    compile_args: Iterable[str] = (),
    exclude_exts: Iterable[str] = (),
    logger=None,
    inplace=None,
    annotate=False,
):
    """Detects pythran extensions under a package and returns a list of
    Extension instances ready to be passed into the ``setup()`` function.

    Parameters
    ----------
    name_package:

        Package to be recursively scanned for Pythran extensions.

    backend : str

        Only initialize extensions for this backend. If None, initialize
        extensions for the default backend (set by an environment variable).

    include_dirs:

        Directories to include while building extensions, for e.g.:
        ``numpy.get_include()``

    compile_args:

        Arguments to be used while compiling extensions

    exclude_ext:

        Extensions to be excluded from the detected list.

    """
    modules = detect_transonic_extensions(name_package, backend)
    if not modules:
        return []

    if backend == "numba":
        # very special case for Numba
        paths = [Path(mod.replace(".", os.path.sep) + ".py") for mod in modules]
        backends["numba"].compile_extensions(paths, None)
        return []
    elif backend == "pythran":
        BackendExtension = PythranExtension
    elif backend == "cython":

        def BackendExtension(mod, files):
            # to avoid a bug: Cython does not take into account .pxd file
            ext = cythonize(files[0], annotate=annotate, language_level=3)[0]
            ext.name = mod
            return ext

    else:
        return []

    if len(exclude_exts) > 0 and logger:
        logger.info(
            "Files in the packages " + str(exclude_exts) + " will not be built."
        )

    if inplace is None:
        inplace = "develop" in sys.argv or (
            "build_ext" in sys.argv and "--inplace" in sys.argv
        )

    extensions = []
    for mod in modules:
        package = mod.rsplit(".", 1)[0]
        if any(package == excluded for excluded in exclude_exts):
            continue
        base_file = mod.replace(".", os.path.sep)
        py_file = base_file + ".py"
        suffix = get_config_var("EXT_SUFFIX")
        bin_file = base_file + suffix
        if (
            not inplace
            or not os.path.exists(bin_file)
            or modification_date(bin_file) < modification_date(py_file)
        ):

            if logger:
                logger.info(
                    "Extension has to be built: {} -> {} ".format(
                        py_file, os.path.basename(bin_file)
                    )
                )

            pext = BackendExtension(mod, [py_file])
            if isinstance(include_dirs, str):
                include_dirs = [include_dirs]
            pext.include_dirs.extend(include_dirs)
            pext.extra_compile_args.extend(compile_args)
            extensions.append(pext)

    return extensions


class ParallelBuildExt(*build_ext_classes):
    @property
    def logger(self):
        logger = get_logger(self.logger_name)
        return logger

    def initialize_options(self):
        """Modify the following to packaging specific needs."""
        super().initialize_options()
        self.logger_name = "transonic"
        self.num_jobs_env_var = ""
        self.ignoreflags = ("-Wstrict-prototypes",)
        self.ignoreflags_startswith = ("-axMIC_", "-diag-disable:")

    def finalize_options(self):
        """Only changed to support setting ``self.parallel`` automatically."""
        if self.parallel is None:
            self.parallel = self.get_num_jobs()

        super().finalize_options()
        self.logger.debug(f"Parallel build enabled with {self.parallel} jobs")
        self.logger.debug(f"Base classes: {build_ext_classes}")

    def get_num_jobs(self):
        try:
            num_jobs = int(os.environ[self.num_jobs_env_var])
        except KeyError:
            import multiprocessing

            num_jobs = multiprocessing.cpu_count()

            try:
                from psutil import virtual_memory
            except ImportError:
                self.logger.warning(
                    "psutil not available at build time. "
                    "Cannot check memory available and potentially limit num_jobs."
                )
            else:
                avail_memory_in_Go = virtual_memory().available / 1e9
                limit_num_jobs = max(1, round(avail_memory_in_Go / 3))
                if num_jobs > limit_num_jobs:
                    self.logger.info(
                        f"num_jobs limited by memory, fixed at {limit_num_jobs}"
                    )
                num_jobs = min(num_jobs, limit_num_jobs)
        return num_jobs

    def build_extensions(self):
        self.check_extensions_list(self.extensions)

        for ext in self.extensions:
            try:
                # For Cython extensions
                ext.sources = self.cython_sources(ext.sources, ext)
            except AttributeError:
                pass

        # Invoke Distutils build_extensions method which respects
        # parallel building. Cython's build_ext ignores this
        SetuptoolsBuildExt.build_extensions(self)

    def _build_extensions_parallel(self):
        """A slightly modified version
        ``setuptools.command.build_ext.build_ext._build_extensions_parallel``
        which:

        - filters out some problematic compiler flags
        - separates extensions of different types into different thread pools.

        """
        logger = self.logger
        if hasattr(self.compiler, "compiler_so"):
            self.compiler.compiler_so = [
                key
                for key in self.compiler.compiler_so
                if key not in self.ignoreflags
                and all(
                    [not key.startswith(s) for s in self.ignoreflags_startswith]
                )
            ]

        # Set of all extension types
        ext_types = {type(ext) for ext in self.extensions}
        extensions_by_type = {T: [] for T in ext_types}
        for ext in self.extensions:
            extensions_by_type[ext.__class__].append(ext)

        def names(exts):
            return [ext.name for ext in exts]

        # Separate building extensions of different types to avoid race conditions
        num_jobs = self.parallel
        logger.info(f"_build_extensions_parallel with num_jobs = {num_jobs}")
        for type_ext, exts in extensions_by_type.items():
            logger.info(
                f"Building extensions of type {type_ext}: {names(exts)}"
            )

            with Pool(num_jobs) as pool:
                pool.map(self.build_extension, exts)

            logger.info(f"Extensions built: {names(exts)}")



"""Numba backend
================

Internal API
------------

.. autoclass:: SubBackendJITNumba
   :members:
   :private-members:

.. autoclass:: NumbaBackend
   :members:
   :private-members:

"""

from typing import Optional

from transonic.analyses.extast import parse, unparse, CommentLine, gast
from transonic.util import format_str

from .py import PythonBackend, SubBackendJITPython


def add_numba_comments(code):
    """Add Numba code in Python comments"""
    mod = parse(code)
    new_body = [CommentLine("# __protected__ from numba import njit")]

    for node in mod.body:
        if isinstance(node, gast.FunctionDef):
            new_body.append(CommentLine("# __protected__ @njit(cache=True, fastmath=True)"))
        new_body.append(node)

    mod.body = new_body
    return format_str(unparse(mod))


class SubBackendJITNumba(SubBackendJITPython):
    def make_backend_source(self, info_analysis, func, path_backend):
        src, has_to_write = super().make_backend_source(
            info_analysis, func, path_backend
        )

        if not src:
            return src, has_to_write

        return add_numba_comments(src), has_to_write


class NumbaBackend(PythonBackend):
    """Main class for the Numba backend"""

    backend_name = "numba"
    _SubBackendJIT = SubBackendJITNumba

    def compile_extension(
        self,
        path_backend,
        name_ext_file=None,
        native=False,
        xsimd=False,
        openmp=False,
        str_accelerator_flags: Optional[str] = None,
        parallel=True,
        force=True,
    ):
        if name_ext_file is None:
            name_ext_file = self.name_ext_from_path_backend(path_backend)

        with open(path_backend) as file:
            source = file.read()

        source = source.replace("# __protected__ ", "")

        with open(path_backend.with_name(name_ext_file), "w") as file:
            file.write(format_str(source))

        compiling = False
        process = None
        return compiling, process

    def _make_backend_code(self, path_py, analyse):
        """Create a backend code from a Python file"""
        code, codes_ext, header = super()._make_backend_code(path_py, analyse)

        if not code:
            return code, codes_ext, header

        return add_numba_comments(code), codes_ext, header



from transonic.typing import format_type_as_backend_type, MemLayout

normalized_types = {"float": "float64", "complex": "complex128", "bool_": "bool"}


class TypeFormatter:
    def __init__(self, backend_name=None):
        self.backend_name = backend_name

    def normalize_type_name(self, name):
        try:
            return normalized_types[name]
        except KeyError:
            return name

    def make_array_code(
        self, dtype, ndim, shape, memview, mem_layout, positive_indices
    ):
        dtype = self.normalize_type_name(dtype.__name__)
        if ndim == 0:
            return dtype
        one_dim = [":"]
        if mem_layout is MemLayout.strided:
            one_dim = ["::"]
        for_shape = one_dim * ndim
        if shape is not None:
            assert ndim == len(shape)
            for index, value in enumerate(shape):
                if value is not None:
                    for_shape[index] = str(value)
        result = f"{dtype}[{', '.join(for_shape)}]"
        if mem_layout is MemLayout.C:
            result += " order(C)"
        elif mem_layout is MemLayout.F:
            result += " order(F)"
        return result

    def make_dict_code(self, type_keys, type_values, **kwargs):
        key = format_type_as_backend_type(type_keys, self, **kwargs)
        value = format_type_as_backend_type(type_values, self, **kwargs)
        return f"{key}: {value} dict"

    def make_set_code(self, type_keys, **kwargs):
        key = format_type_as_backend_type(type_keys, self, **kwargs)
        return f"{key} set"

    def make_list_code(self, type_elem, **kwargs):
        return format_type_as_backend_type(type_elem, self, **kwargs) + " list"

    def make_tuple_code(self, types, **kwargs):
        strings = [
            format_type_as_backend_type(type_, self, **kwargs) for type_ in types
        ]
        return f"({', '.join(strings)})"

    def make_const_code(self, code):
        return code


base_type_formatter = TypeFormatter()



"""Configuration
================

Transonic is sensible to the environment variables:

- :code:`TRANSONIC_DIR` can be set to control where the cached files are
  saved.

- :code:`TRANSONIC_COMPILE_AT_IMPORT` can be set to enable a mode for which
  Transonic compiles at import time the Pythran file associated with the
  imported module. This behavior can also be triggered programmatically by using
  the function :code:`set_compile_at_import`.

- :code:`TRANSONIC_NO_REPLACE` can be set to disable all code replacements.
  This is useful only when measuring code coverage.

- :code:`FLUID_COMPILE_JIT` can be set to false to disable the
  compilation of jited functions. This can be useful for unittests.

- :code:`TRANSONIC_MPI_TIMEOUT` sets the MPI timeout (default to 5 s).

By the way, for performance, it is important to configure Pythran with a file
`~/.pythranrc
<https://pythran.readthedocs.io/en/latest/MANUAL.html#customizing-your-pythranrc>`_:

For Linux, I use:

.. code::

    [pythran]
    complex_hook = True

    [compiler]
    CXX=clang++-6.0
    CC=clang-6.0
    blas=openblas

User API
--------

.. autofunction:: set_backend

"""

import os
from pathlib import Path
from warnings import warn

path_root = Path(os.environ.get("TRANSONIC_DIR", Path.home() / ".transonic"))


def strtobool(value):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    value = value.lower()
    if value in ("y", "yes", "t", "true", "on", "1"):
        return 1
    elif value in ("n", "no", "f", "false", "off", "0"):
        return 0
    else:
        raise ValueError(f"invalid truth value {value}")


has_to_replace = not strtobool(os.environ.get("TRANSONIC_NO_REPLACE", "0"))

backend_default = "pythran"
backend_set_by_user = False


def set_backend(backend: str):
    """Set the "global variable" backend_default"""

    backend = backend.lower()
    supported_backends = ["pythran", "cython", "numba", "python"]
    if backend not in supported_backends:
        raise ValueError(f"backend {backend} not supported")

    global backend_default, backend_set_by_user

    backend_default = backend
    backend_set_by_user = True

    # warning: this import has to be here!
    from transonic.util import can_import_accelerator

    if not can_import_accelerator(backend):
        warn(f'Backend set to "{backend}" but accelerator not importable')


try:
    _backend = os.environ["TRANSONIC_BACKEND"]
except KeyError:
    pass
else:
    set_backend(_backend)



"""Capture the external nodes used in functions
===============================================
"""

import gast as ast

from transonic.analyses import beniget
from transonic.analyses import extast


class CaptureX(ast.NodeVisitor):
    """Capture the external nodes used in functions, classes and blocks"""

    def __init__(
        self,
        functions,
        module_node,
        ancestors=None,
        defuse_chains=None,
        usedef_chains=None,
        consider_annotations=True,
        blocks=None,
    ):

        if defuse_chains is None:
            self.du_chains = du = beniget.DefUseChains()
            du.visit(module_node)
            self.ud_chains = beniget.UseDefChains(du)
            self.ancestors = beniget.Ancestors()
            self.ancestors.visit(module_node)
        else:
            self.du_chains = defuse_chains
            self.ud_chains = usedef_chains
            self.ancestors = ancestors

        self.consider_annotations = consider_annotations
        self._annot = None

        self.external = []
        self.visited_external = set()

        self.functions = functions
        for func in functions:
            self.func = func
            self.visit(func)

        if blocks is None:
            return

        for nodes in blocks:
            if nodes:
                node = nodes[0]
                self.func = ancestors.parentFunction(node)
                for node in nodes:
                    self.visit(node)

    def visit_Name(self, node):

        parent_node = self.ancestors.parents(node)[-1]
        if (
            isinstance(parent_node, ast.FunctionDef)
            and node == parent_node.returns
            and not self.consider_annotations
        ):
            return

        # register load of identifiers not locally defined
        if isinstance(node.ctx, ast.Load):
            try:
                self.ud_chains.chains[node]
            except KeyError:
                if not (
                    isinstance(parent_node, ast.FunctionDef)
                    and node == parent_node.returns
                ):
                    raise
                from warnings import warn

                warn(f"BUG Beniget (node.id={node.id}), but we try to continue!")
                return

            for def_ in self.ud_chains.chains[node]:
                try:
                    parents = self.ancestors.parents(def_.node)
                except KeyError:
                    return  # a builtin
                if self.func not in parents:
                    if isinstance(def_.node, ast.FunctionDef):
                        defining_node = def_.node
                    else:
                        defining_node = self.ancestors.parentStmt(def_.node)
                    if defining_node not in self.visited_external:
                        self.rec(defining_node)
                        if defining_node in self.functions:
                            return

                        if (
                            self.consider_annotations == "only"
                            and self._annot is None
                        ):
                            return

                        if defining_node not in self.visited_external:
                            self.visited_external.add(defining_node)
                            self.external.append(defining_node)
        elif (
            isinstance(node.ctx, (ast.Param, ast.Store))
            and self.consider_annotations
        ):
            if node.annotation is None:
                return
            self._annot = node.annotation
            self.visit(node.annotation)
            self._annot = None

    def visit_ClassDef(self, node_class):
        for node in node_class.body:
            if isinstance(node, ast.AnnAssign):
                self._annot = node.annotation
                self.visit(node)
                self._annot = None

    def visit_AnnAssign(self, node):
        if self.consider_annotations:
            self._annot = node.annotation
            self.visit(node.annotation)
            self._annot = None

        if node.value is not None and self.consider_annotations != "only":
            self.visit(node.value)

    def rec(self, node):
        "walk definitions to find their operands's def"
        if node is self.func:
            return

        if isinstance(node, ast.Assign):
            self.visit(node.value)
        elif isinstance(node, ast.FunctionDef):
            # tmp: to deal with @include
            node.decorator_list = []

            old_func = self.func
            self.func = node
            self.visit(node)
            self.func = old_func

        # TODO: implement this for AugAssign etc

    def make_code_external(self):
        code = []
        for node in self.external:
            code.append(extast.unparse(node).strip())
        return "\n".join(code)


if __name__ == "__main__":

    code = "a = 1; b = [a, a]\ndef foo():\n return b"
    # code = "a = 1; b = len([a, a])\ndef foo():\n return b"
    # code = "import numpy as np\na = np.int(1)\ndef foo():\n return np.zeros(a)"

    code = """

a = 1

def fooo():
    return 1

def foo():
    return a + fooo()

def bar():
    return foo()

    """

    module = extast.parse(code)
    function = module.body[3]
    capturex = CaptureX((function,), module)

    print(capturex.make_code_external())



"""Parser using tokenize
========================

"""

from tokenize import tokenize, COMMENT
from io import BytesIO


def parse_transonic_def_commands(code: str):
    """Parse the code of a .py file and return data"""

    functions = set()
    signatures_func = {}

    in_def = False

    g = tokenize(BytesIO(code.encode("utf-8")).readline)
    for toknum, tokval, a, b, c in g:
        if toknum == COMMENT:
            if tokval.startswith("#transonic"):
                tokval = "# " + tokval[1:]

            if tokval.startswith("# transonic def "):
                in_def = True
                signature_func = tokval.split("# transonic def ", 1)[1]
                name_func = signature_func.split("(")[0]
                functions.add(name_func)

                if name_func not in signatures_func:
                    signatures_func[name_func] = []

                if ")" in tokval:
                    in_def = False
                    signatures_func[name_func].append(signature_func)

            if in_def:
                if "# transonic def " in tokval:
                    tokval = tokval.split("(", 1)[1]
                signature_func += tokval.replace("#", "").strip()
                if ")" in tokval:
                    in_def = False
                    signatures_func[name_func].append(signature_func)

    return signatures_func



"""Command line transonic
============================

Internal API
------------

.. autofunction:: run

.. autofunction:: parse_args

"""

import argparse
from pathlib import Path
from glob import glob
import sys

from transonic import __version__

from transonic.compiler import wait_for_all_extensions, scheduler

from .backends import backends
from transonic.config import backend_default
from transonic.log import logger
from transonic.util import (
    has_to_build,
    clear_cached_extensions,
    can_import_accelerator,
)

doc = """
transonic: easily speedup your Python code with Pythran

"""


def run():
    """Run the transonic commandline

    See :code:`transonic -h`
    """
    args = parse_args()

    if args.version:
        print(__version__)
        return

    if not args.path and not args.clear_cache:
        logger.warning("No python files given. Nothing to do! ✨ 🍰 ✨.")
        return

    if args.clear_cache:
        clear_cached_extensions(args.clear_cache, args.force, args.backend)
        return

    if args.verbose is None:
        logger.set_level(None)
    elif args.verbose == 1:
        logger.set_level("info")
    elif args.verbose > 1:
        logger.set_level("debug")
    logger.info(args)

    path = args.path

    if isinstance(path, list) and len(path) == 1:
        path = path[0]

    if isinstance(path, list):
        paths = path
    else:
        path = Path(path)
        if path.is_file():
            paths = (path,)
        elif path.is_dir():
            paths = path.glob("*.py")
        else:
            paths = glob(str(path))

    if not paths:
        logger.error(f"No input file found (args.path = {args.path})")
        sys.exit(1)

    backend = backends[args.backend]
    backend.make_backend_files(paths, force=args.force)

    if args.no_compile:
        return

    if not can_import_accelerator(backend.name):
        logger.warning(
            f"Since {backend.name_capitalized} is not importable, "
            "Transonic cannot properly compile a file."
        )
        return

    # find pythran files not already compiled
    backends_paths = []

    for path in paths:
        path = Path(path)
        backend_path = path.parent / str(f"__{backend.name}__") / path.name
        ext_path = backend_path.with_name(
            backend.name_ext_from_path_backend(backend_path)
        )
        if backend_path.exists() and has_to_build(ext_path, backend_path):
            backends_paths.append(backend_path)

    with scheduler.progress:
        backend.compile_extensions(
            backends_paths,
            str_accelerator_flags=args.accelerator_flags,
            parallel=True,
            force=args.force,
        )

        if not args.no_blocking:
            wait_for_all_extensions()


def parse_args():
    """Parse the arguments"""
    parser = argparse.ArgumentParser(
        description=doc, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("path", help="Path file or directory.", nargs="*")

    parser.add_argument(
        "-f",
        "--force",
        help="proceed even if the files seem up-to-date",
        action="store_true",
    )

    parser.add_argument(
        "-V", "--version", help="print version and exit", action="store_true"
    )

    parser.add_argument("-v", "--verbose", help="verbose mode", action="count")

    parser.add_argument(
        "-b",
        "--backend",
        help=("Backend (pythran, cython, numba or python)"),
        type=str,
        default=backend_default,
    )

    parser.add_argument(
        "-nc",
        "--no-compile",
        help="do not compile the Pythran/Cython/... files",
        action="store_true",
    )

    parser.add_argument(
        "-nb",
        "--no-blocking",
        help="launch the compilation in the background and return",
        action="store_true",
    )

    parser.add_argument(
        "-pf",
        "--pythran-flags",
        help=("Depreciated: use -af"),
        type=str,
        default="",
    )

    parser.add_argument(
        "-af",
        "--accelerator-flags",
        help=(
            "Flags sent to the accelerator. "
            'Default is "". '
            "There has to be atleast one space in the passed string! "
            "Examples:\n"
            '`transonic foo.py -af "-march=native "` or\n'
            '`transonic foo.py -af "-march=native -DUSE_XSIMD -Ofast"`\n'
        ),
        type=str,
        default="",
    )

    parser.add_argument(
        "-cc",
        "--clear-cache",
        help=("clear the cached extensions for a module"),
        type=str,
        # default="",
    )

    args = parser.parse_args()
    if args.pythran_flags != "":
        raise DeprecationWarning("-pf is deprecated. Use -af instead!")

    return args


if __name__ == "__main__":
    run()



"""Internal utilities
=====================

Public API
----------

.. autofunction:: set_compile_at_import

Internal API
------------

.. autofunction:: find_module_name_from_path

.. autofunction:: get_module_name

.. autofunction:: modification_date

.. autofunction:: has_to_build

.. autofunction:: get_source_without_decorator

.. autoclass:: TypeHintRemover
   :members:
   :private-members:

.. autofunction:: strip_typehints

.. autofunction:: get_ipython_input

.. autofunction:: get_info_from_ipython

.. autofunction:: has_to_compile_at_import

.. autofunction:: import_from_path

.. autofunction:: query_yes_no

.. autofunction:: clear_cached_extensions

.. autofunction:: is_method

.. autofunction:: has_to_write

.. autofunction:: write_if_has_to_write

"""

import os
import sys
import inspect
import re
from pathlib import Path
import importlib.util
import shutil
from textwrap import dedent
from typing import Callable

import gast as ast

from transonic.config import backend_default

try:
    # since black is still beta (in 03/2019), we cannot impose a version :-(
    import black
except ImportError:
    import autopep8

    def format_str(src_contents):
        return autopep8.fix_code(src_contents)


else:
    try:
        _mode = black.FileMode(line_length=82)
    except TypeError:

        def format_str(src_contents: str):
            try:
                return black.format_str(src_contents, line_length=82)
            except black.InvalidInput:
                print("black.InvalidInput\n" + src_contents)
                raise

    else:

        def format_str(src_contents: str):
            try:
                return black.format_str(src_contents, mode=_mode)
            except black.InvalidInput:
                print("black.InvalidInput\n" + src_contents)
                raise


try:
    from IPython.core.getipython import get_ipython
except ImportError:
    pass

from transonic import __version__
from transonic.analyses import extast

from transonic.compiler import (
    ext_suffix,
    make_hex,
    modification_date,
    has_to_build,
)

from transonic.config import path_root, strtobool


__all__ = ["modification_date", "has_to_build", "path_root"]


def can_import_accelerator(backend: str = backend_default):
    if backend == "pythran":
        try:
            import pythran
        except ImportError:
            return False
    elif backend == "cython":
        try:
            import cython
        except ImportError:
            return False
    elif backend == "numba":
        try:
            import numba
        except ImportError:
            return False
    elif backend == "python":
        return True
    else:
        raise NotImplementedError
    return True


def print_versions(accelerators=None):

    print(f"Transonic {__version__}")

    if accelerators is None or "pythran" in accelerators:
        try:
            import pythran
        except ImportError:
            print("Pythran not importable")
        else:
            print(f"Pythran {pythran.__version__}")

    if accelerators is None or "numba" in accelerators:
        try:
            import numba
        except ImportError:
            print("Numba not importable")
        else:
            print(f"Numba {numba.__version__}")

    if accelerators is None or "cython" in accelerators:
        try:
            import Cython
        except ImportError:
            print("Cython not importable")
        else:
            print(f"Cython {Cython.__version__}")


def find_module_name_from_path(path_py: Path):
    """Find the module name from the path of a Python file

    It is done by looking to ``sys.path`` to see how the module can be imported.

    """
    path_py = Path(path_py)

    cwd = Path.cwd()
    path = path_py.absolute().parent
    module_name = path_py.stem

    # special case for jit_classes
    special_dir = "__jit_class__"
    if special_dir in path.parts:
        tmp = [special_dir]
        name_pack = ".".join(path.parts[path.parts.index(special_dir) + 1 :])
        if name_pack:
            tmp.append(name_pack)
        tmp.append(module_name)
        return ".".join(tmp)

    while path.parents:
        if path == cwd or str(path) in sys.path:
            return module_name

        module_name = path.name + "." + module_name
        path = path.parent

    return path_py.stem

def _get_filename_from_frame(frame):
    return frame.f_code.co_filename


def get_module_name(frame):
    """Get the full module name"""
    module = inspect.getmodule(frame)
    filename = None
    if module is not None:
        module_name = module.__name__
        if module_name in ("__main__", "<run_path>"):
            filename = _get_filename_from_frame(frame)
    else:
        filename = _get_filename_from_frame(frame)

    if filename is not None:
        module_name = find_module_name_from_path(Path(filename))

    if module_name is None:
        # ipython ?
        src, module_name = get_info_from_ipython()

    return module_name


def get_name_calling_module():
    return get_module_name(get_frame(1))


def get_frame(depth=1):
    """Return a frame object from the call stack.

    This should be equivalent to ``inspect.currentframe().f_back`` with ``depth``
    times ``.f_back``. For example, ``depth = 1`` gives the caller frame.
    """
    try:
        return sys._getframe(depth+1)
    except AttributeError:
        # we might want to implement this with another function
        # (``inspect.currentframe`` or ``inspect.stack()``)
        # for alternative Python implementations
        raise
    except ValueError:
        print([frame_info[1] for frame_info in inspect.stack()])
        raise


def get_source_without_decorator(func: Callable):
    """Get the source of a function without its decorator"""
    src = inspect.getsource(func)
    src = dedent(src)
    return strip_typehints(re.sub(r"@.*?\sdef\s", "def ", src))


class TypeHintRemover(ast.NodeTransformer):
    """Strip the type hints

    from https://stackoverflow.com/a/42734810/1779806
    """

    def visit_FunctionDef(self, fdef):
        # remove the return type defintion
        fdef.returns = None
        # remove all argument annotations
        if fdef.args.args:
            for arg in fdef.args.args:
                arg.annotation = None

        body = []
        for node in fdef.body:
            if isinstance(node, ast.AnnAssign):
                if node.value is None:
                    continue
                node = ast.Assign(
                    targets=[node.target], value=node.value, type_comment=None
                )
            body.append(node)
        fdef.body = body

        return fdef


def strip_typehints(source):
    """Strip the type hints from a function"""
    source = format_str(source)
    # parse the source code into an AST
    parsed_source = ast.parse(source)
    # remove all type annotations, function return type definitions
    # and import statements from 'typing'
    transformed = TypeHintRemover().visit(parsed_source)
    # convert the AST back to source code
    striped_code = extast.unparse(transformed)
    return striped_code


def make_code_from_fdef_node(fdef):
    transformed = TypeHintRemover().visit(fdef)
    # convert the AST back to source code
    code = extast.unparse(transformed)
    return format_str(code)


def get_ipython_input(last=True):
    """Get the input code when called from IPython"""
    ip = get_ipython()

    hist_raw = ip.history_manager.input_hist_raw
    if last:
        return hist_raw[-1]
    else:
        return "\n".join(hist_raw)


def get_info_from_ipython():
    """Get the input code and a "filename" when called from IPython"""
    src = get_ipython_input()
    hex_input = make_hex(src)
    dummy_filename = "__ipython__" + hex_input
    return src, dummy_filename


_PYTHRANIZE_AT_IMPORT = None


def set_compile_at_import(value=True):
    """Control the "compile_at_import" mode"""
    global _PYTHRANIZE_AT_IMPORT
    _PYTHRANIZE_AT_IMPORT = value


def has_to_compile_at_import():
    """Check if transonic has to pythranize at import time"""
    if _PYTHRANIZE_AT_IMPORT is not None:
        return _PYTHRANIZE_AT_IMPORT
    return "TRANSONIC_COMPILE_AT_IMPORT" in os.environ


def import_from_path(path: Path, module_name: str):
    """Import a .py file or an extension from its path"""
    if not path.exists():
        raise ImportError(
            f"File {path} does not exist. "
            f"[path.name for path in path.parent.glob('*')]:\n{[path.name for path in path.parent.glob('*')]}\n"
        )

    if "." in module_name:
        package_name, mod_name = module_name.rsplit(".", 1)
        name_file = path.name.split(".", 1)[0]
        if mod_name != name_file:
            module_name = ".".join((package_name, name_file))
    else:
        module_name = path.stem

    if module_name in sys.modules:
        module = sys.modules[module_name]

        if module.__file__.endswith(ext_suffix) and Path(module.__file__) == path:
            return module

    spec = importlib.util.spec_from_file_location(module_name, path)
    # for potential "local imports" in the module
    sys.path.insert(0, str(path.parent))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    # clean sys.path
    sys.path.pop(0)
    # fix bug Numba
    sys.modules[module_name] = module
    return module


def query_yes_no(question: str, default: str = None, force: bool = False):
    """User yes or no query"""
    if force:
        return True

    if default is None:
        end = "(y/n)"
        default = ""
    elif default == "y":
        end = "([y]/n)"
    elif default == "n":
        end = "(y/[n])"

    print(f"{question} {end}")
    while True:
        answer = input()
        if answer == "":
            answer = default
        try:
            return strtobool(answer)
        except ValueError:
            print('Please respond with "y" or "n".')


def clear_cached_extensions(module_name: str, force: bool, backend: str):
    """Delete the cached extensions related to a module"""

    from transonic.backends import backends
    from transonic import mpi

    backend = backends[backend]
    path_jit = mpi.Path(backend.jit.path_base)

    if module_name.endswith(".py"):
        module_name = module_name[:-3]

    if os.path.sep not in module_name:
        relative_path = module_name.replace(".", os.path.sep)
    else:
        relative_path = module_name

    path_pythran_dir_jit = path_jit / relative_path

    relative_path = Path(relative_path)

    path_pythran = relative_path.parent / (
        "__{backend.name}__/" + relative_path.name + ".py"
    )
    path_ext = path_pythran.with_name(
        backend.name_ext_from_path_backend(path_pythran)
    )

    if not path_pythran_dir_jit.exists() and not (
        path_pythran.exists() or path_ext.exists()
    ):
        print(
            f"Not able to find cached extensions corresponding to {module_name}"
        )
        print("Nothing to do! ✨ 🍰 ✨.")
        return

    if path_pythran_dir_jit.exists() and query_yes_no(
        f"Do you confirm that you want to delete the cached files for {module_name}",
        default="y",
        force=force,
    ):
        print(f"Remove directory {path_pythran_dir_jit}")
        shutil.rmtree(path_pythran_dir_jit)

    if path_pythran.exists() or path_ext.exists():
        if query_yes_no(
            f"Do you confirm that you want to delete the AOT cache for {module_name}",
            default="y",
            force=force,
        ):
            for path in (path_pythran, path_ext):
                if path.exists():
                    path.unlink()


def is_method(func):
    """Determine wether a function is going to be used as a method"""
    signature = inspect.signature(func)
    try:
        answer = next(iter(signature.parameters.keys())) == "self"
    except StopIteration:
        answer = False

    return answer


def has_to_write(path_file: Path, new_code: str):
    """Check if a file exists and contains a code"""
    if path_file.exists():
        with open(path_file, "r") as file:
            source_code = file.read()
        if new_code == source_code:
            return False
        else:
            return True
    else:
        return True


def write_if_has_to_write(
    path_file: Path, new_code: str, logger=None, force=False
):
    """Write a file if it doesn't exist or doesn't contain a particular code"""
    written = False
    if has_to_write(path_file, new_code) or force:
        written = True
        path_file.parent.mkdir(exist_ok=True, parents=True)
        with open(path_file, "w") as file:
            file.write(new_code)
        if logger:
            logger(f"{path_file} written")
    return written


def timeit(stmt="pass", setup="pass", total_duration=2, globals=None):
    """timeit with a approximate total_duration"""
    from timeit import Timer

    timer = Timer(stmt, setup, globals=globals)
    timer.timeit(10)
    duration1 = timer.timeit(10) / 10
    number = max(1, int(round(total_duration / duration1)))

    if number > 200:
        repeat = int(round(number / 100))
        number = int(round(number / repeat))
    else:
        repeat = 1

    if number * duration1 > 2 * total_duration:
        raise RuntimeError("number * duration1 > 2 * total_duration")

    duration = min(timer.repeat(repeat=repeat, number=number))
    return duration / number


def timeit_verbose(
    stmt,
    setup="pass",
    total_duration=2,
    globals=None,
    norm=None,
    name=None,
    print_time=False,
    max_length_name=33,
):
    result = timeit(
        stmt, setup=setup, total_duration=total_duration, globals=globals
    )
    if norm is None:
        norm = result
        norm_given = False
    else:
        norm_given = True

    if name is None:
        name = stmt.split("(")[0]

    fmt_name = f"{{:{max_length_name}s}}"
    name = fmt_name.format(name)

    if print_time:
        raw_time = f" = {result:7.3g} s"
    else:
        raw_time = ""

    print(f"{name}: {result/norm:5.3g} * norm{raw_time}")
    if not norm_given and not print_time:
        print(f"norm = {norm:5.3g} s")

    return result
